
/**
 * CollisionForce - DESCRIPTION
 * <p/>
 * An extension to the Processing project - http://Proce55ing.net
 * <p/>
 * Copyright (c) 2004 Simon Greenwold
 * Created: Feb 11, 2004
 * <p/>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <p/>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU Lesser General
 * Public License along with the Processing project; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

public class CollisionForce extends Force {
    Particle a;

    float radius;
    float minStrength;
    float maxStrength;

    CollisionForce(Particle a) {
      this.a = a;
    }

    float falloff(float dist) {
      if (dist == 0) return 1;
      return  1.0f / dist;
    }

    void applyForce() {
      Particle p;
      Particle[] particles = a.pSystem.particles;
      for (int i = 0; i < a.pSystem.nParticles; i++) {
        p = particles[i];
        applyForce(p);
      }
    }

    void applyForce(Particle p) {
      if (a == p) return;
      if (p.fixed() ||
          p.collisionForce == null ||
          !p.collisionForce.active ||
          a.collisionRadius + p.collisionRadius <= 0) return;

      minStrength = a.minCollisionForce;
      maxStrength = a.maxCollisionForce;
      radius = a.collisionRadius;

      dist = a.dist(a.pos, p.pos);
      float distAdj = (dist - radius) - p.collisionRadius;

      if (distAdj > 0) return;

      float ratio = -distAdj / (radius + p.collisionRadius);
      float finalStrength = minStrength + (ratio * (maxStrength - minStrength));

      Lx = a.pos[0] - p.pos[0];
      Ly = a.pos[1] - p.pos[1];
      Lz = a.pos[2] - p.pos[2];
      Fx = -finalStrength * Lx / dist;
      Fy = -finalStrength * Ly / dist;
      Fz = -finalStrength * Lz / dist;

      p.addForce(Fx, Fy, Fz);
    }

    float dist;
    float Fx;
    float Fy;
    float Fz;
    float Lx;
    float Ly;
    float Lz;
}
