
import java.util.Iterator;
import java.util.ArrayList;

/**
 * Magnet - DESCRIPTION
 * <p/>
 * An extension to the Processing project - http://Proce55ing.net
 * <p/>
 * Copyright (c) 2004 Simon Greenwold
 * Created: Feb 11, 2004
 * <p/>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <p/>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU Lesser General
 * Public License along with the Processing project; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

public class Magnet extends Force {
    Particle a;

    float maxForce;
    int type;
    float strength;
    ArrayList ignoreTypes = new ArrayList();
  private PSystem pSystem;

    Magnet() {
      super();
      type = 0;
      maxForce = 1;
      strength = 1;
    }

    Magnet(Particle a) {
      this();
      this.a = a;
    }

  void init(PSystem pSystem) {
    this.pSystem = pSystem;
  }

    void ignoreType(Particle particleType) {
      if (ignoreTypes.contains(particleType.getClass())) return;
      ignoreTypes.add(particleType.getClass());
    }

    float falloff(float dist) {
      if (dist == 0) return 1;
      return  1.0f / dist;
    }

    void applyForce() {
      Particle p;
      Particle particles[] = pSystem.particles;
      for (int i = 0; i < pSystem.nParticles; i++) {
        p = particles[i];
        applyForce(p);
      }
    }

    void applyForce(Particle p) {
      if (a == p) return;
      if (p.fixed()) return;

      for (Iterator iterator = ignoreTypes.iterator(); iterator.hasNext();) {
        Class aClass = (Class) iterator.next();
        if (aClass.isInstance(p)) {
          return;
        }
      }

      dist = a.dist(a.pos, p.pos);
      Lx = a.pos[0] - p.pos[0];
      Ly = a.pos[1] - p.pos[1];
      Lz = a.pos[2] - p.pos[2];
      falloffStrength = a.constrain(strength * (falloff(dist)), -a.abs(maxForce), a.abs(maxForce));
      Fx = falloffStrength * Lx / dist;
      Fy = falloffStrength * Ly / dist;
      Fz = falloffStrength * Lz / dist;

      p.addForce(Fx, Fy, Fz);
    }

    float dist;
    float Fx;
    float Fy;
    float Fz;
    float Lx;
    float Ly;
    float Lz;
    float falloffStrength;

}
