
/**
 * Spring - DESCRIPTION
 * <p/>
 * An extension to the Processing project - http://Proce55ing.net
 * <p/>
 * Copyright (c) 2004 Simon Greenwold
 * Created: Feb 11, 2004
 * <p/>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <p/>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU Lesser General
 * Public License along with the Processing project; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

public class Spring extends Force {

    Particle a;
    Particle b;
    private PSystem pSystem;

    float restLength;  // Rest length
    float strength; // Spring constant
    float damping; // Damping constant

    void draw() {
        if (pSystem == null) return;
        pSystem.getBApplet().line(a.pos[0], a.pos[1], a.pos[2], b.pos[0], b.pos[1], b.pos[2]);
    }

    Spring(Particle a, Particle b) {
        this();
        this.a = a;
        this.b = b;
    }

    Spring() {
    }

    void init(PSystem pSystem) {
        this.pSystem = pSystem;
        strength = pSystem.defaultSpringStrength;
        damping = pSystem.defaultSpringDamping;
        restLength = pSystem.defaultSpringRestLength;
    }

    void applyForce() {
        dist = a.dist(a.pos, b.pos);
        if (dist == 0) {
            return;
        }
        Lx = a.pos[0] - b.pos[0];
        Ly = a.pos[1] - b.pos[1];
        Lz = a.pos[2] - b.pos[2];

        Fx = - (strength * (dist - restLength) + damping * (a.velocity[0] - b.velocity[0]) * Lx / dist) * Lx / dist;
        Fy = - (strength * (dist - restLength) + damping * (a.velocity[1] - b.velocity[1]) * Ly / dist) * Ly / dist;
        Fz = - (strength * (dist - restLength) + damping * (a.velocity[2] - b.velocity[2]) * Lz / dist) * Lz / dist;

        a.addForce(Fx, Fy, Fz);
        b.addForce(-Fx, -Fy, -Fz);
    }

    int timeCounter = 0;

    void applyForce(Particle p) {
        if (a != p && b != p) return;

        dist = a.dist(a.pos, b.pos);
        if (dist == 0) {
            return;
        }
        Lx = a.pos[0] - b.pos[0];
        Ly = a.pos[1] - b.pos[1];
        Lz = a.pos[2] - b.pos[2];
        Fx = - (strength * (dist - restLength) + damping * (a.velocity[0] - b.velocity[0]) * Lx / dist) * Lx / dist;
        Fy = - (strength * (dist - restLength) + damping * (a.velocity[1] - b.velocity[1]) * Ly / dist) * Ly / dist;
        Fz = - (strength * (dist - restLength) + damping * (a.velocity[2] - b.velocity[2]) * Lz / dist) * Lz / dist;

        /*if (timeCounter > 1010) {
        System.out.println("Report particle " + timeCounter + " " + this);
        System.out.println("  " + Fx + " " + Fy + " " + Fz + " " + dist);
        System.out.println("  " + Lx + " " + Ly + " " + Lz);
        System.out.println("  " + strength + " " + restLength + " " + damping);
        System.out.println("  " + a + " " + a.velocity[0] + " " + a.velocity[1] + " " + a.velocity[2]);
        System.out.println("  " + b + " " + b.velocity[0] + " " + b.velocity[1] + " " + b.velocity[2]);
        }  */

        timeCounter++;

        if (a == p)
            a.addForce(Fx, Fy, Fz);
        else
            b.addForce(-Fx, -Fy, -Fz);
    }

    float dist;
    float Fx;
    float Fy;
    float Fz;
    float Lx;
    float Ly;
    float Lz;
}
