// Alpha Mask
// by toxi <http://www.toxi.co.uk>

// Loads a "mask" for an image which specifies its transparency 
// in different parts of the image. The two images are blended
// together using the blend() function.

// Created 29 April 2003

AlphaImg  testImg;

void setup() 
{
  size(200,200);
  testImg = new AlphaImg("test.jpg", "mask.jpg");
}

void loop() 
{
  background((mouseX+mouseY)/1.5);
  testImg.display(width/2,height/2);
  testImg.display(mouseX,mouseY);
}

class AlphaImg 
{
  BImage  img;
  int[]   mask;
  int     area;
    
  AlphaImg(String basefile, String maskfile) 
  {
    img=loadImage(basefile);
    BImage alpha=loadImage(maskfile);
    area=img.width*img.height;
    setAlpha(alpha.pixels);
  }
  
  // Overwrite current alpha channel if the array contains RGB data, 
  // Only the blue channel will be used as alpha data
  void setAlpha(int[] alpha) 
  {
    if (alpha.length==area) {
      mask=new int[area];
      for(int i=0; i<area; i++) {
        mask[i]=alpha[i]&0xff;
      }
    }
  }
  
  // Return dimensions of the image
  int getWidth() { 
    return img.width; 
  }
  int getHeight() { 
    return img.height;
  }
  
  // Generic clipping
  int low(int a, int b) { 
    return (a<b) ? a : b; 
  }
  
  int high(int a, int b) { 
    return (a>b) ? a : b; 
  }

  // Generic linear interpolation
  int mix(int a,int b, int f) 
  { 
    return a+(((b-a)*f)>>8); 
  }
  
  // Blend 2 colours
  int blend(int a, int b, int f) 
  {
    return mix(a>>16,b>>16,f)<<16 | mix(a>>8&0xff,b>>8&0xff,f)<<8 | mix(a&0xff,b&0xff,f);
  }

  // Main display method
  void display(int x, int y) 
  {
    x-=img.width/2; 
    y-=img.height/2;
    if (-x>=img.width || -y>=img.height) return;
    int tw=(x<0) ? x+img.width : min(width-x,img.width);
    int th=(y<0) ? y+img.height : min(height-y, img.height);
    int offset1=((x<0) ? -x : 0)+((y<0) ? -y : 0)*img.width;
    int offset2=((x<0) ? 0 : x)+((y<0) ? 0 : y)*width;
    
    int twidth1=img.width-tw;
    int twidth2=width-tw;
    
    int yy=-1;

    while(++yy<th) {
     int xx=-1;
     while(++xx<tw) {
       pixels[offset2] = blend(pixels[offset2], img.pixels[offset1], mask[offset1]);
       offset1++; 
       offset2++;
     }
     offset1+=twidth1; offset2+=twidth2;
     }
  }
  
}





